import copy
import unittest

import torch
import torch.nn as nn

from bnn.engine import BinaryChef
from bnn.ops import (
    BasicInputBinarizer,
    BasicScaleBinarizer,
    XNORWeightBinarizer
)


class Flatten(nn.Module):
    def __init__(self) -> None:
        super(Flatten, self).__init__()

    def forward(self, x):
        return x.view(x.size(0), -1)


class EngineTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.net = nn.Sequential(
            nn.Conv2d(3, 16, 1, 1),
            nn.BatchNorm2d(16),
            nn.ReLU(inplace=True),
            nn.Conv2d(16, 16, 1, 1),
            nn.BatchNorm2d(16),
            nn.ReLU(inplace=True),
            nn.AdaptiveAvgPool2d((1, 1)),
            Flatten(),
            nn.Linear(16, 3)
        )

        self.asset_path = 'test/assets/test.yaml'

    def test_step_length(self):
        chef = BinaryChef(self.asset_path)

        self.assertEqual(len(chef), 3)

    def test_engine_function(self):
        model = copy.copy(self.net)
        chef = BinaryChef(self.asset_path)

        # Step1
        model = chef.next(model)
        # self.assertFalse(hasattr(model[0], 'bconfig'))
        self.assertTrue(hasattr(model[3], 'bconfig'))
        self.assertIsInstance(model[3].weight_pre_process, nn.Identity)

        # Step2
        model = chef.next(model)
        # self.assertFalse(hasattr(model[0], 'bconfig'))
        self.assertTrue(hasattr(model[3], 'bconfig'))
        self.assertIsInstance(model[3].weight_pre_process, XNORWeightBinarizer)
        w_alpha = model[3].activation_post_process.alpha

        # Step3
        model = chef.next(model)
        # self.assertTrue(hasattr(model[0], 'bconfig'))
        self.assertTrue(hasattr(model[3], 'bconfig'))
        self.assertIsInstance(model[3].weight_pre_process, XNORWeightBinarizer)
        self.assertTrue(torch.equal(w_alpha, model[3].activation_post_process.alpha))


if __name__ == '__main__':
    unittest.main()



import copy
import unittest

import torch
import torch.nn as nn

from bnn import BConfig, prepare_binary_model
from bnn.layers import Conv2d, Linear
from bnn.ops import (
    BasicInputBinarizer,
    BasicScaleBinarizer,
    XNORWeightBinarizer
)


class Flatten(nn.Module):
    def __init__(self) -> None:
        super(Flatten, self).__init__()

    def forward(self, x):
        return x.view(x.size(0), -1)


class BinarizerTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.linear_layer = nn.Linear(10, 3)
        self.conv_layer = nn.Conv2d(3, 16, 1, 1)
        self.net = nn.Sequential(
            nn.Conv2d(3, 16, 1, 1),
            nn.BatchNorm2d(16),
            nn.ReLU(inplace=True),
            nn.Conv2d(16, 16, 1, 1),
            nn.BatchNorm2d(16),
            nn.ReLU(inplace=True),
            nn.AdaptiveAvgPool2d((1, 1)),
            Flatten(),
            nn.Linear(16, 3)
        )

        self.input1 = torch.rand(1, 10)
        self.input2 = torch.rand(1, 3, 8, 8)

        self.random_bconfig = BConfig(
            activation_pre_process=BasicInputBinarizer,
            activation_post_process=BasicScaleBinarizer,
            weight_pre_process=XNORWeightBinarizer
        )

    def tearDown(self) -> None:
        pass

    def weight_reset(self, m):
        if isinstance(m, nn.Conv2d) or isinstance(m, nn.Linear):
            m.reset_parameters()

    def test_single_linear_layer(self):
        model = copy.copy(self.linear_layer)
        model = prepare_binary_model(model, bconfig=self.random_bconfig)

        self.assertEqual(type(model), Linear)

    def test_single_conv2d_layer(self):
        model = copy.copy(self.conv_layer)
        model = prepare_binary_model(model, bconfig=self.random_bconfig)

        self.assertEqual(type(model), Conv2d)

    def test_many_layers(self):
        model = copy.copy(self.linear_layer)
        model = prepare_binary_model(model, bconfig=self.random_bconfig)

        self.assertEqual(type(model), Linear)

    def test_skip_binarization(self):
        model = copy.copy(self.net)

        fp32_config = BConfig(
            activation_pre_process=nn.Identity,
            activation_post_process=nn.Identity,
            weight_pre_process=nn.Identity
        )
        model = prepare_binary_model(model, bconfig=self.random_bconfig, custom_config_layers_name={'8': fp32_config})

        cnt_conv, cnt_linear = 0, 0
        for module in model.modules():
            if isinstance(module, Conv2d):
                cnt_conv += 1
            elif isinstance(module, Linear):
                if isinstance(module.activation_pre_process, nn.Identity):
                    cnt_linear += 1

        self.assertEqual(cnt_conv, 2)
        self.assertEqual(cnt_linear, 1)

    def test_save_load_state_dict(self):
        model = copy.deepcopy(self.net)
        x = self.input2.clone()

        model = prepare_binary_model(model, bconfig=self.random_bconfig)
        out1 = model(x)

        binary_state_dict = model.state_dict()

        model = copy.deepcopy(self.net)
        model.apply(self.weight_reset)
        model = prepare_binary_model(model, bconfig=self.random_bconfig)
        model.load_state_dict(binary_state_dict)
        out2 = model(x)

        self.assertTrue(torch.equal(out1, out2))


class OpsTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.input = torch.tensor([0.3, 0.1, -2, -0.001, 0.01])
        self.conv_layer = nn.Conv2d(3, 16, 1, 1)

    def test_basic_input_binarizer(self):
        funct = BasicInputBinarizer()
        self.assertTrue(torch.equal(funct(self.input.clone()), torch.sign(self.input.clone())))

    def test_BasicScaleBinarizer(self):
        funct = BasicScaleBinarizer(copy.copy(self.conv_layer))

    def test_XNORWeightBinarizer(self):
        funct = XNORWeightBinarizer()

if __name__ == '__main__':
    unittest.main()



from functools import partial
from abc import ABCMeta, abstractmethod
import math
import torch
import torch.nn as nn
import torch.nn.functional as F
from typing import Dict, Any, List


def _with_args(cls_or_self: Any, **kwargs: Dict[str, Any]) -> Any:
    r"""Wrapper that allows creation of class factories.
    This can be useful when there is a need to create classes with the same
    constructor arguments, but different instances.
    Source: https://github.com/pytorch/pytorch/blob/b02c932fb67717cb26d6258908541b670faa4e72/torch/quantization/observer.py
    Example::
        >>> Foo.with_args = classmethod(_with_args)
        >>> foo_builder = Foo.with_args(a=3, b=4).with_args(answer=42)
        >>> foo_instance1 = foo_builder()
        >>> foo_instance2 = foo_builder()
        >>> id(foo_instance1) == id(foo_instance2)
        False
    """
    class _PartialWrapper(object):
        def __init__(self, p):
            self.p = p

        def __call__(self, *args, **keywords):
            return self.p(*args, **keywords)

        def __repr__(self):
            return self.p.__repr__()

        with_args = _with_args
    r = _PartialWrapper(partial(cls_or_self, **kwargs))
    return r

ABC = ABCMeta(str("ABC"), (object,), {})


class BinarizerBase(ABC, nn.Module):
    def __init__(self) -> None:
        super(BinarizerBase, self).__init__()

    @abstractmethod
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        pass

    with_args = classmethod(_with_args)


class SignActivation(torch.autograd.Function):
    r"""Applies the sign function element-wise
    :math:`\text{sgn(x)} = \begin{cases} -1 & \text{if } x < 0, \\ 1 & \text{if} x >0  \end{cases}`
    the gradients of which are computed using a STE, namely using :math:`\text{hardtanh(x)}`.
    Shape:
        - Input: :math:`(N, *)` where `*` means, any number of additional
          dimensions
        - Output: :math:`(N, *)`, same shape as the input
    Examples::
        >>> input = torch.randn(3)
        >>> output = SignActivation.apply(input)
    """
    @staticmethod
    def forward(ctx, input: torch.Tensor) -> torch.Tensor:
        ctx.save_for_backward(input)
        return input.sign()

    @staticmethod
    def backward(ctx, grad_output: torch.Tensor) -> torch.Tensor:
        input, = ctx.saved_tensors
        grad_input = grad_output.clone()
        grad_input.masked_fill_(input.ge(1) | input.le(-1), 0)
        return grad_input


class SignActivationStochastic(SignActivation):
    r"""Binarize the data using a stochastic binarizer
    :math:`\text{sgn(x)} = \begin{cases} -1 & \text{with probablity } p = \sigma(x), \\ 1 & \text{with probablity } 1 - p \end{cases}`
    the gradients of which are computed using a STE, namely using :math:`\text{hardtanh(x)}`.
    Shape:
        - Input: :math:`(N, *)` where `*` means, any number of additional
          dimensions
        - Output: :math:`(N, *)`, same shape as the input
    Examples::
        >>> input = torch.randn(3)
        >>> output = SignActivationStochastic.apply(input)
    """
    @staticmethod
    def forward(ctx, input: torch.Tensor) -> torch.Tensor:
        ctx.save_for_backward(input)
        noise = torch.rand_like(input).sub_(0.5)
        return input.add_(1).div_(2).add_(noise).clamp_(0, 1).round_().mul_(2).sub_(1)


class XNORWeightBinarizer(BinarizerBase):
    r"""Binarize the parameters of a given layer using the analytical solution
    proposed in the XNOR-Net paper.
    :math:`\text{out} = \frac{1}{n}\norm{\mathbf{W}}_{\ell} \text{sgn(x)}(\mathbf{W})`
    Shape:
        - Input: :math:`(N, *)` where `*` means, any number of additional
          dimensions
        - Output: :math:`(N, *)`, same shape as the input
    Examples::
        >>> binarizer = XNORWeightBinarizer()
        >>> output = F.conv2d(input, binarizer(weight))
    Args:
        compute_alpha: if True, compute the real-valued scaling factor
        center_weights: make the weights zero-mean
    """

    def __init__(self, compute_alpha: bool = True, center_weights: bool = False) -> None:
        super(XNORWeightBinarizer, self).__init__()
        self.compute_alpha = compute_alpha
        self.center_weights = center_weights

    def _compute_alpha(self, x: torch.Tensor) -> torch.Tensor:
        n = x[0].nelement()
        if x.dim() == 4:
            alpha = x.norm(1, 3, keepdim=True).sum([2, 1], keepdim=True).div_(n)
        elif x.dim() == 2:
            alpha = x.norm(1, 1, keepdim=True).div_(n)
        else:
            raise ValueError(f"Expected ndims equal with 2 or 4, but found {x.dim()}")

        return alpha

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.center_weights:
            mean = x.mean(1, keepdim=True).expand_as(x)
            x = x.sub(mean)

        if self.compute_alpha:
            alpha = self._compute_alpha(x)
            x = SignActivation.apply(x).mul_(alpha.expand_as(x))
        else:
            x = SignActivation.apply(x)

        return x


class BasicInputBinarizer(BinarizerBase):
    r"""Applies the sign function element-wise.
    nn.Module version of SignActivation.
    """

    def __init__(self):
        super(BasicInputBinarizer, self).__init__()

    def forward(self, x: torch.Tensor) -> None:
        return SignActivation.apply(x)


class StochasticInputBinarizer(BinarizerBase):
    r"""Applies the sign function element-wise.
    nn.Module version of SignActivation.
    """

    def __init__(self):
        super(StochasticInputBinarizer, self).__init__()

    def forward(self, x: torch.Tensor):
        return SignActivationStochastic.apply(x)


class AdvancedInputBinarizer(BinarizerBase):
    def __init__(self, derivative_funct=torch.tanh, t: int = 5):
        super(AdvancedInputBinarizer, self).__init__()
        self.derivative_funct = derivative_funct
        self.t = t

    def forward(self, x: torch.tensor) -> torch.Tensor:
        x = self.derivative_funct(x * self.t)
        with torch.no_grad():
            x = torch.sign(x)
        return x


class BasicScaleBinarizer(BinarizerBase):
    def __init__(self, module: nn.Module, shape: List[int] = None) -> None:
        super(BasicScaleBinarizer, self).__init__()

        if isinstance(module, nn.Linear):
            num_channels = module.out_features
        elif isinstance(module, nn.Conv2d):
            num_channels = module.out_channels
        else:
            if hasattr(module, 'out_channels'):
                num_channels = module.out_channels
            else:
                raise Exception('Unknown layer of type {} missing out_channels'.format(type(module)))

        if shape is None:
            alpha_shape = [1, num_channels] + [1] * (module.weight.dim() - 2)
        else:
            alpha_shape = shape
        self.alpha = nn.Parameter(torch.ones(*alpha_shape))

    def forward(self, layer_out: torch.Tensor, layer_in: torch.Tensor) -> torch.Tensor:
        x = layer_out
        return x.mul_(self.alpha)

    def extra_repr(self) -> str:
        return '{}'.format(list(self.alpha.size()))


class XNORScaleBinarizer(BinarizerBase):
    def __init__(self, module: nn.Module) -> None:
        super(BasicScaleBinarizer, self).__init__()
        kernel_size = module.kernel_size
        self.stride = module.stride
        self.padding = module.padding
        self.register_buffer('fixed_weight', torch.ones(*kernel_size).div_(math.prod(kernel_size)), persistent=False)

    def forward(self, layer_out: torch.Tensor, layer_in: torch.Tensor) -> torch.Tensor:
        x = layer_out
        scale = torch.mean(dim=1, keepdim=True)
        scale = F.conv2d(scale, self.fixed_weight, stride=self.stride, padding=self.padding)

        return x.mul_(scale)



from dataclasses import fields
import torch

from .. import BConfig


def copy_paramters(source_mod: torch.nn.Module, target_mod: torch.nn.Module, bconfig: BConfig) -> None:
    attributes = [field.name for field in fields(bconfig)]
    for attribute in attributes:
        attr_obj_source = getattr(source_mod, attribute, None)
        attr_obj_target = getattr(target_mod, attribute, None)
        for name, source_param in attr_obj_source.named_parameters():
            if source_param is not None and attr_obj_target is not None:
                target_param = getattr(attr_obj_target, name, None)
                if target_param is not None:
                    if torch.equal(torch.tensor(target_param.size()), torch.tensor(source_param.size())):
                        target_param.data.copy_(source_param.data)



import yaml
import re
from easydict import EasyDict as edict
from typing import List, Callable, Dict

from . import BConfig, prepare_binary_model
from .ops import *


def _option_builder_helper(partial_config: Dict[str, str]) -> str:
    if 'args' in partial_config:
        start_string = '.with_args('
        end_string = ')'
        content = ''
        for k, v in partial_config.args.items():
            content += '{}={},'.format(k, v)
        final_string = start_string + content + end_string

        return final_string
    return ''


class BinaryChef(object):
    r"""Converts a given model according to the configutation and steps defined in an YAML file.ut
    Examples::
        >>> bc = BinaryChef('config.yaml')
        >>> for i in range(len(bc)):
        >>>     model = bc.next(model)
        >>>     # do training logic for desired number of epochs
    Args:
        config: path to a valid yaml file containing the steps
        user_modules: list containing custom user defined binarizers
    """

    def __init__(self, config: str, user_modules: List[Callable[..., nn.Module]] = []) -> None:
        with open(config) as file:
            config = yaml.load(file, Loader=yaml.FullLoader)
            config = [edict(config[k]) for k in config.keys()]
        self.config = config
        self.current_step = 0
        # inject the received functions into the namespace
        for user_module in user_modules:
            globals()[user_module.__name__] = user_module

    def __len__(self) -> int:
        return len(self.config)

    def get_num_steps(self) -> int:
        return len(self)

    def run_step(self, model: nn.Module, step: int) -> nn.Module:
        assert len(self) > step
        step_config = self.config[step]

        # Ignore certain white listed layers
        print(step_config, type(step_config))
        ignore_layer_names = step_config.ignore_layer_names if 'ignore_layer_names' in step_config else []

        # prepare model
        bconfig = BConfig(
            activation_pre_process=eval(
                step_config.pre_activation.name
                + _option_builder_helper(
                    step_config.pre_activation)),
            activation_post_process=eval(
                step_config.post_activation.name
                + _option_builder_helper(
                    step_config.post_activation)),
            weight_pre_process=eval(
                step_config.weight.name
                + _option_builder_helper(
                    step_config.weight)))
        bmodel = prepare_binary_model(model, bconfig=bconfig, ignore_layers_name=ignore_layer_names)

        return bmodel

    def next(self, model: nn.Module) -> nn.Module:
        self.current_step += 1
        return self.run_step(model, self.current_step - 1)



from dataclasses import asdict
from .bconfig import BConfig
import re
import copy
import logging
import torch
import torch.nn as nn
from typing import Dict, List

from . import layers as bnn

DEFAULT_MODULE_MAPPING = {
    nn.Linear: bnn.Linear,
    nn.Conv2d: bnn.Conv2d,
    nn.Conv1d: bnn.Conv1d
}
for k, v in copy.copy(DEFAULT_MODULE_MAPPING).items():
    DEFAULT_MODULE_MAPPING[v] = v


def _get_first_layer(model: nn.Module) -> List[str]:
    for name, module in model.named_modules():
        if type(module) in DEFAULT_MODULE_MAPPING:
            return [name]
    return []


def _get_last_layer(model: nn.Module) -> List[str]:
    for name, module in reversed(list(model.named_modules())):
        if type(module) in DEFAULT_MODULE_MAPPING:
            return [name]
    return []


def _regex_match(model: nn.Module, pattern: str, modules_mapping: Dict[nn.Module, nn.Module]) -> List[str]:
    # Strip first and last character (expected to be $ and $)
    pattern = pattern[1:-1]

    matched_names = []
    pattern = re.compile(pattern)
    for name, module in model.named_modules():
        if type(module) in modules_mapping:
            if pattern.search(name) is not None:
                matched_names.append(name)
    return matched_names

_KNOWN_SPECIAL_WORDS = {
    '_last_': _get_first_layer,
    '_first_': _get_last_layer
}


def get_unique_devices_(module: nn.Module) -> List[torch.device]:
    return {p.device for p in module.parameters()} | \
        {p.device for p in module.buffers()}


def get_modules_to_binarize(model: nn.Module, bconfig: BConfig,
                            modules_mapping: Dict[nn.Module, nn.Module] = None,
                            custom_config_layers_name: Dict[str, BConfig] = {},
                            ignore_layers_name: List[str] = []) -> Dict[str, nn.Module]:
    if modules_mapping is None:
        modules_mapping = DEFAULT_MODULE_MAPPING

    # Parse special cases
    processed_ignore_layer_names = []
    for name in ignore_layers_name:
        if name in _KNOWN_SPECIAL_WORDS.keys():
            processed_ignore_layer_names += _KNOWN_SPECIAL_WORDS[name](model)
        elif name[0] == '$' and name[-1] == '$':
            processed_ignore_layer_names += _regex_match(model, name, modules_mapping)
        else:
            processed_ignore_layer_names.append(name)

    modules_to_replace = {}
    for name, module in model.named_modules():
        if type(module) in modules_mapping:
            if name in processed_ignore_layer_names:
                continue

            layer_config = copy.copy(bconfig)
            # Use layer specific customization
            if name in custom_config_layers_name:
                for k, v in asdict(custom_config_layers_name[name]).items():
                    setattr(layer_config, k, v)

            # respect device affinity when swapping modules
            devices = get_unique_devices_(module)
            assert len(devices) <= 1, (
                "swap_module only works with cpu or single-device CUDA modules, "
                "but got devices {}".format(devices)
            )
            device = next(iter(devices)) if len(devices) > 0 else None

            modules_to_replace[name] = modules_mapping[type(module)].from_module(
                module,
                layer_config
            )
            if device:
                modules_to_replace[name].to(device)
        elif name in custom_config_layers_name:
            logging.warning('Module named {} defined in the configuration was not found.'.format(name))
    return modules_to_replace


def swap_modules_by_name(model: nn.Module, modules_to_replace: Dict[str, nn.Module],
                         modules_mapping: Dict[nn.Module, nn.Module] = None) -> nn.Module:
    if modules_mapping is None:
        modules_mapping = DEFAULT_MODULE_MAPPING

    def _swap_module(module: nn.Module):
        for name, child in module.named_children():
            if type(child) in modules_mapping:
                for n, m in model.named_modules():
                    if child is m:
                        if n in modules_to_replace:
                            setattr(module, name, modules_to_replace.pop(n))
                        break
            else:
                _swap_module(child)
    if len(list(model.named_children())) == 0:
        if type(model) in modules_mapping and len(modules_to_replace.keys()) == 1:
            model = modules_to_replace[list(modules_to_replace.keys())[0]]
    else:
        _swap_module(model)
    return model


def prepare_binary_model(model: nn.Module, bconfig: BConfig,
                         modules_mapping: Dict[nn.Module, nn.Module] = None,
                         custom_config_layers_name: Dict[str, BConfig] = {},
                         ignore_layers_name: List[str] = []) -> nn.Module:
    modules_to_replace = get_modules_to_binarize(
        model,
        bconfig,
        modules_mapping,
        custom_config_layers_name,
        ignore_layers_name)
    model = swap_modules_by_name(model, modules_to_replace, modules_mapping)

    return model



from typing import Tuple, List

import torch
import torch.nn as nn

from .layers.bats_ops import *


class Cell(nn.Module):

    def __init__(
            self,
            genotype: Genotype,
            C_prev_prev: int,
            C_prev: int,
            C: int,
            reduction: bool,
            reduction_prev: bool,
            groups: int = 12,
            use_shake_shake: bool = False) -> None:
        super(Cell, self).__init__()
        self.use_shake_shake = use_shake_shake
        if reduction_prev:
            self.preprocess0 = FactorizedReduce(C_prev_prev, C)
        else:
            self.preprocess0 = ReLUConvBN(C_prev_prev, C, 1, 1, 0)
        self.preprocess1 = ReLUConvBN(C_prev, C, 1, 1, 0)
        if reduction:
            op_names, indices = zip(*genotype.reduce)
            concat = genotype.reduce_concat
        else:
            op_names, indices = zip(*genotype.normal)
            concat = genotype.normal_concat
        self._compile(C, op_names, indices, concat, reduction, groups)

    def _compile(
            self,
            C: int,
            op_names: List[str],
            indices: List[int],
            concat: List[int],
            reduction: bool,
            groups: int) -> None:
        assert len(op_names) == len(indices)
        self._steps = len(op_names) // 2
        self._concat = concat
        self.multiplier = len(concat)

        self._ops = nn.ModuleList()
        for name, index in zip(op_names, indices):
            stride = 2 if reduction and index < 2 else 1
            op = OPS[name](C, stride, True, True, groups)
            self._ops += [op]
        self._indices = indices

    def forward(self, s0: torch.Tensor, s1: torch.Tensor, drop_prob: float) -> torch.Tensor:
        s0 = self.preprocess0(s0)
        s1 = self.preprocess1(s1)

        states = [s0, s1]
        for i in range(self._steps):
            h1 = states[self._indices[2 * i]]
            h2 = states[self._indices[2 * i + 1]]
            op1 = self._ops[2 * i]
            op2 = self._ops[2 * i + 1]
            h1 = op1(h1)
            h2 = op2(h2)
            if self.training and drop_prob > 0.:
                if not isinstance(op1, nn.Identity):
                    h1 = drop_path(h1, drop_prob)
                if not isinstance(op2, nn.Identity):
                    h2 = drop_path(h2, drop_prob)
            s = h1 + h2
            states += [s]

        if self.use_shake_shake:
            if self.training:
                shake = torch.softmax(torch.zeros(len(self._concat)).uniform_(), dim=0)
                return torch.cat([states[i] * shake[j].item() for j, i in enumerate(self._concat)], dim=1)
            else:
                return torch.cat([states[i] * (1 / len(self._concat)) for i in self._concat], dim=1)
        else:
            return torch.cat([states[i] for i in self._concat], dim=1)


class AuxiliaryHead(nn.Module):

    def __init__(self, C: int, num_classes: int, stride: int) -> None:
        super(AuxiliaryHead, self).__init__()
        self.features = nn.Sequential(
            nn.AvgPool2d(5, stride=stride, padding=0, count_include_pad=False),
            nn.BatchNorm2d(C),
            nn.Conv2d(C, 128, 1, bias=False),
            nn.PReLU(num_parameters=128),

            nn.BatchNorm2d(128),
            nn.Conv2d(128, 768, 2, bias=False),
            nn.PReLU(num_parameters=768),
        )
        self.classifier = nn.Linear(768, num_classes)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.features(x)
        x = self.classifier(x.view(x.size(0), -1))
        return x


class BATSNetworkCIFAR(nn.Module):

    def __init__(self, C: int, num_classes: int, layers: int, auxiliary: bool, genotype, groups: int) -> None:
        super(BATSNetworkCIFAR, self).__init__()
        self._layers = layers
        self._auxiliary = auxiliary
        stem_multiplier = 3
        C_curr = stem_multiplier * C
        self.stem = nn.Sequential(
            nn.Conv2d(3, C_curr, 3, padding=1, bias=False),
            nn.BatchNorm2d(C_curr),
            nn.ReLU(inplace=True),
        )
        C_prev_prev, C_prev, C_curr = C_curr, C_curr, C
        self.cells = nn.ModuleList()
        reduction_prev = False
        for i in range(layers):
            if i in [layers // 3, 2 * layers // 3]:
                C_curr *= 2
                reduction = True
            else:
                reduction = False
            cell = Cell(genotype, C_prev_prev, C_prev, C_curr, reduction, reduction_prev, groups)
            reduction_prev = reduction
            self.cells += [cell]
            C_prev_prev, C_prev = C_prev, cell.multiplier * C_curr
            if i == 2 * layers // 3:
                C_to_auxiliary = C_prev
        if auxiliary:
            self.auxiliary_head = AuxiliaryHead(C_to_auxiliary, num_classes, 3)
        self.global_pooling = nn.AdaptiveAvgPool2d(1)
        self.classifier = nn.Linear(C_prev, num_classes)

    def forward(self, input: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        logits_aux = None
        s0 = s1 = self.stem(input)
        for i, cell in enumerate(self.cells):
            s0, s1 = s1, cell(s0, s1, self.drop_path_prob)
            if i == 2 * self._layers // 3:
                if self._auxiliary and self.training:
                    logits_aux = self.auxiliary_head(s1)
        out = self.global_pooling(s1)
        logits = self.classifier(out.view(out.size(0), -1))
        return logits, logits_aux


class BATSNetworkImageNet(nn.Module):

    def __init__(self, C: int, num_classes: int, layers: int, auxiliary: bool, genotype, groups: int) -> None:
        super(BATSNetworkImageNet, self).__init__()
        self._layers = layers
        self._auxiliary = auxiliary

        self.stem0 = nn.Sequential(
            nn.Conv2d(3, C // 2, kernel_size=3, stride=2, padding=1, bias=False),
            nn.BatchNorm2d(C // 2),
            nn.ReLU(inplace=True),
            nn.Conv2d(C // 2, C, 3, stride=2, padding=1, bias=False, groups=C // 20),
            nn.BatchNorm2d(C),
        )
        self.stem1 = nn.Sequential(
            nn.ReLU(inplace=True),
            nn.Conv2d(C, C, 3, stride=2, padding=1, bias=False, groups=C // 20),
            nn.BatchNorm2d(C),
        )
        C_prev_prev, C_prev, C_curr = C, C, C

        self.cells = nn.ModuleList()
        reduction_prev = True
        for i in range(layers):
            if i in [layers // 3, 2 * layers // 3]:
                C_curr *= 2
                reduction = True
            else:
                reduction = False
            cell = Cell(genotype, C_prev_prev, C_prev, C_curr, reduction, reduction_prev, groups)
            reduction_prev = reduction
            self.cells += [cell]
            C_prev_prev, C_prev = C_prev, cell.multiplier * C_curr
            if i == 2 * layers // 3:
                C_to_auxiliary = C_prev

        if auxiliary:
            self.auxiliary_head = AuxiliaryHead(C_to_auxiliary, num_classes, 2)
        self.global_pooling = nn.AvgPool2d(7)
        self.classifier = nn.Linear(C_prev, num_classes)

    def forward(self, input: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        logits_aux = None
        s0 = self.stem0(input)
        s1 = self.stem1(s0)
        for i, cell in enumerate(self.cells):
            s0, s1 = s1, cell(s0, s1, self.drop_path_prob)
            if i == 2 * self._layers // 3:
                if self._auxiliary and self.training:
                    logits_aux = self.auxiliary_head(s1)
        out = self.global_pooling(s1)
        logits = self.classifier(out.view(out.size(0), -1))
        return logits, logits_aux



from dataclasses import dataclass
import torch
import torch.nn as nn


class Identity(nn.Identity):
    def forward(self, layer_out: torch.Tensor, layer_in: torch.Tensor) -> torch.Tensor:
        return layer_out


@dataclass
class BConfig:
    activation_pre_process: nn.Module = nn.Identity
    activation_post_process: nn.Module = Identity
    weight_pre_process: nn.Module = nn.Identity

    def __post_init__(self) -> None:
        if isinstance(
                self.activation_pre_process,
                nn.Module) or isinstance(
                self.activation_post_process,
                nn.Module) or isinstance(
                self.weight_pre_process,
                nn.Module):
            raise ValueError("BConfig received an instance, please pass the class instead.")



import torch.nn as nn


def conv3x3(in_planes: int, out_planes: int, stride: int = 1, groups: int = 1, dilation: int = 1
            ) -> nn.Module:
    """3x3 convolution with padding"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                     padding=dilation, groups=groups, bias=False, dilation=dilation)


def conv1x1(in_planes: int, out_planes: int, stride: int = 1
            ) -> nn.Module:
    """1x1 convolution"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=False)



from typing import Callable, Optional
import torch
import torch.nn as nn

from .common import conv1x1, conv3x3


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self,
                 inplanes: int,
                 planes: int,
                 stride: int = 1,
                 downsample: Optional[nn.Module] = None,
                 groups: int = 1,
                 base_width: int = 64,
                 dilation: int = 1,
                 norm_layer: Optional[Callable[..., nn.Module]] = None,
                 activation=nn.ReLU
                 ) -> None:
        super(BasicBlock, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if groups != 1 or base_width != 64:
            raise ValueError('BasicBlock only supports groups=1 and base_width=64')
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in BasicBlock")
        # Both self.conv1 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = norm_layer(planes)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = norm_layer(planes)
        self.act1 = activation(inpace=True) if activation == nn.ReLU else activation(num_parameters=planes)
        self.act2 = activation(inpace=True) if activation == nn.ReLU else activation(num_parameters=planes)

        self.downsample = downsample
        self.stride = stride

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.act1(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.act2(out)

        return out


class Bottleneck(nn.Module):
    # Bottleneck in torchvision places the stride for downsampling at 3x3 convolution(self.conv2)
    # while original implementation places the stride at the first 1x1 convolution(self.conv1)
    # according to "Deep residual learning for image recognition"https://arxiv.org/abs/1512.03385.
    # This variant is also known as ResNet V1.5 and improves accuracy according to
    # https://ngc.nvidia.com/catalog/model-scripts/nvidia:resnet_50_v1_5_for_pytorch.

    expansion = 4

    def __init__(self,
                 inplanes: int,
                 planes: int,
                 stride: int = 1,
                 downsample: Optional[nn.Module] = None,
                 groups: int = 1,
                 base_width: int = 64,
                 dilation: int = 1,
                 norm_layer: Optional[Callable[..., nn.Module]] = None,
                 activation=nn.ReLU
                 ) -> None:
        super(Bottleneck, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        width = int(planes * (base_width / 64.)) * groups
        # Both self.conv2 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv1x1(inplanes, width)
        self.bn1 = norm_layer(width)
        self.conv2 = conv3x3(width, width, stride, groups, dilation)
        self.bn2 = norm_layer(width)
        self.conv3 = conv1x1(width, planes * self.expansion)
        self.bn3 = norm_layer(planes * self.expansion)
        self.act1 = activation(inpace=True) if activation == nn.ReLU else activation(num_parameters=width)
        self.act2 = activation(inpace=True) if activation == nn.ReLU else activation(num_parameters=width)
        self.act3 = activation(
            inpace=True) if activation == nn.ReLU else activation(
            num_parameters=planes * self.expansion)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.act1(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.act2(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.act3(out)

        return out


class PreBasicBlock(nn.Module):
    expansion = 1

    def __init__(self,
                 inplanes: int,
                 planes: int,
                 stride: int = 1,
                 downsample: Optional[nn.Module] = None,
                 groups: int = 1,
                 base_width: int = 64,
                 dilation: int = 1,
                 norm_layer: Optional[Callable[..., nn.Module]] = None,
                 activation=nn.ReLU
                 ) -> None:
        super(PreBasicBlock, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if groups != 1 or base_width != 64:
            raise ValueError('BasicBlock only supports groups=1 and base_width=64')
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in BasicBlock")
        # Both self.conv1 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = norm_layer(inplanes)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = norm_layer(planes)
        self.act1 = activation(inpace=True) if activation == nn.ReLU else activation(num_parameters=planes)
        self.act2 = activation(inpace=True) if activation == nn.ReLU else activation(num_parameters=planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.bn1(x)
        out = self.conv1(out)
        out = self.act1(out)

        out = self.bn2(out)
        out = self.conv2(out)
        out = self.act2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        return out


class PreBottleneck(nn.Module):
    # Bottleneck in torchvision places the stride for downsampling at 3x3 convolution(self.conv2)
    # while original implementation places the stride at the first 1x1 convolution(self.conv1)
    # according to "Deep residual learning for image recognition"https://arxiv.org/abs/1512.03385.
    # This variant is also known as ResNet V1.5 and improves accuracy according to
    # https://ngc.nvidia.com/catalog/model-scripts/nvidia:resnet_50_v1_5_for_pytorch.

    expansion = 4

    def __init__(self,
                 inplanes: int,
                 planes: int,
                 stride: int = 1,
                 downsample: Optional[nn.Module] = None,
                 groups: int = 1,
                 base_width: int = 64,
                 dilation: int = 1,
                 norm_layer: Optional[Callable[..., nn.Module]] = None,
                 activation=nn.ReLU
                 ) -> None:
        super(PreBottleneck, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        width = int(planes * (base_width / 64.)) * groups
        # Both self.conv2 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv1x1(inplanes, width)
        self.bn1 = norm_layer(inplanes)
        self.conv2 = conv3x3(width, width, stride, groups, dilation)
        self.bn2 = norm_layer(width)
        self.conv3 = conv1x1(width, planes * self.expansion)
        self.bn3 = norm_layer(width)
        self.act1 = activation(inpace=True) if activation == nn.ReLU else activation(num_parameters=width)
        self.act2 = activation(inpace=True) if activation == nn.ReLU else activation(num_parameters=width)
        self.act3 = activation(
            inpace=True) if activation == nn.ReLU else activation(
            num_parameters=planes * self.expansion)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.bn1(x)
        out = self.conv1(out)
        out = self.act1(out)

        out = self.bn2(out)
        out = self.conv2(out)
        out = self.act2(out)

        out = self.bn3(out)
        out = self.conv3(out)
        out = self.act3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity

        return out



import torch
import torch.nn as nn

from collections import namedtuple

Genotype = namedtuple('Genotype', 'normal normal_concat reduce reduce_concat')

PRIMITIVES = [
    'none',
    'max_pool_3x3',
    'avg_pool_3x3',
    'skip_connect',
    'sep_conv_3x3',
    'sep_conv_5x5',
    'dil_conv_3x3',
    'dil_conv_5x5'
]


def channel_shuffle(x: torch.Tensor, groups: int) -> torch.Tensor:
    batchsize, num_channels, height, width = x.data.size()

    channels_per_group = num_channels // groups

    x = x.view(batchsize, groups, channels_per_group, height, width)
    x = torch.transpose(x, 1, 2).contiguous()

    x = x.view(batchsize, -1, height, width)

    return x


def drop_path(x: torch.Tensor, drop_prob: float) -> torch.Tensor:
    if drop_prob > 0.:
        keep_prob = 1. - drop_prob
        mask = torch.tensor(x.size(0), 1, 1, 1, device=x.device).bernoulli_(keep_prob)
        x.div_(keep_prob)
        x.mul_(mask)
    return x

OPS = {
    'none': lambda C, stride, affine, skip, groups: Zero(stride),
    'avg_pool_3x3': lambda C, stride, affine, skip, groups: nn.AvgPool2d(3, stride=stride, padding=1, count_include_pad=False),
    'max_pool_3x3': lambda C, stride, affine, skip, groups: nn.MaxPool2d(3, stride=stride, padding=1),
    'skip_connect': lambda C, stride, affine, skip, groups: nn.Identity() if stride == 1 else FactorizedReduce(C, C, affine=affine),
    'sep_conv_3x3': lambda C, stride, affine, skip, groups: SepConv(C, C, 3, stride, 1, affine=affine, skip=skip, groups=groups),
    'sep_conv_5x5': lambda C, stride, affine, skip, groups: SepConv(C, C, 5, stride, 2, affine=affine, skip=skip, groups=groups),
    'sep_conv_7x7': lambda C, stride, affine, skip, groups: SepConv(C, C, 7, stride, 3, affine=affine, skip=skip, groups=groups),
    'dil_conv_3x3': lambda C, stride, affine, skip, groups: DilConv(C, C, 3, stride, 2, 2, affine=affine, skip=skip, groups=groups),
    'dil_conv_5x5': lambda C, stride, affine, skip, groups: DilConv(C, C, 5, stride, 4, 2, affine=affine, skip=skip, groups=groups),
    'conv_7x1_1x7': lambda C, stride, affine, skip, groups: FactorizedConv(C, 7, stride, affine=affine, skip=skip),
}


class FactorizedConv(nn.Module):
    def __init__(self, C: int, kernel_size: int, stride: int, affine: bool = True, skip: bool = False) -> None:
        super(FactorizedConv, self).__init__()
        self.skip = skip or True
        self.stride = stride
        self.op = nn.Sequential(
            nn.BatchNorm2d(C, affine=affine),
            nn.Conv2d(C, C, (1, kernel_size), stride=(1, stride), padding=(0, kernel_size // 2), bias=False),
            nn.PReLU(num_parameters=C),

            nn.BatchNorm2d(C, affine=affine),

            nn.Conv2d(C, C, (kernel_size, 1), stride=(stride, 1), padding=(kernel_size // 2, 0), bias=False),
            nn.PReLU(num_parameters=C),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.skip and self.stride == 1:
            return x + channel_shuffle(self.op(x), 4)
        else:
            return channel_shuffle(self.op(x), 4)


class ReLUConvBN(nn.Module):

    def __init__(
            self,
            C_in: int,
            C_out: int,
            kernel_size: int,
            stride: int,
            padding: int,
            affine: bool = True,
            skip: bool = False) -> None:
        super(ReLUConvBN, self).__init__()
        self.skip = skip or True
        self.stride = stride
        self.C_in = C_in
        self.C_out = C_out

        self.op = nn.Sequential(
            nn.BatchNorm2d(C_in, affine=affine),
            nn.Conv2d(C_in, C_out, kernel_size, stride=stride, padding=padding, bias=False),
            nn.PReLU(num_parameters=C_out),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.skip and self.stride == 1 and self.C_in == self.C_out:
            return x + self.op(x)
        else:
            return self.op(x)


class DilConv(nn.Module):

    def __init__(
            self,
            C_in: int,
            C_out: int,
            kernel_size: int,
            stride: int,
            padding: int,
            dilation: int,
            affine: bool = True,
            skip: bool = False,
            groups: int = 12) -> None:
        super(DilConv, self).__init__()
        self.skip = skip or True
        self.stride = stride
        self.op = nn.Sequential(
            nn.BatchNorm2d(
                C_in,
                affine=affine),
            nn.Conv2d(
                C_in,
                C_in,
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
                dilation=dilation,
                groups=groups,
                bias=False),
            nn.PReLU(
                num_parameters=C_in),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.skip and self.stride == 1:
            return x + channel_shuffle(self.op(x), 4)
        else:
            return channel_shuffle(self.op(x), 4)


class SepConv(nn.Module):

    def __init__(
            self,
            C_in: int,
            C_out: int,
            kernel_size: int,
            stride: int,
            padding: int,
            affine: bool = True,
            skip: bool = False,
            groups: int = 12) -> None:
        super(SepConv, self).__init__()
        self.skip = skip or True
        self.stride = stride
        self.op = nn.Sequential(
            nn.BatchNorm2d(C_in, affine=affine),
            nn.Conv2d(C_in, C_in, kernel_size=kernel_size, stride=stride, padding=padding, groups=groups, bias=False),
            nn.PReLU(num_parameters=C_in),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.skip and self.stride == 1:
            return x + channel_shuffle(self.op(x), 4)
        else:
            return channel_shuffle(self.op(x), 4)


class Zero(nn.Module):

    def __init__(self, stride: int) -> None:
        super(Zero, self).__init__()
        self.stride = stride

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        n, c, h, w = x.size()
        h //= self.stride
        w //= self.stride
        padding = torch.zeros(n, c, h, w, dtype=torch.float32, device=x.device)
        return padding


class FactorizedReduce(nn.Module):

    def __init__(self, C_in: int, C_out: int, affine: bool = True) -> None:
        super(FactorizedReduce, self).__init__()
        assert C_out % 2 == 0
        self.activation = nn.PReLU(num_parameters=C_out)
        self.conv_1 = nn.Sequential(
            nn.Conv2d(C_in, C_out // 2, 1, stride=2, padding=0, bias=False)
        )
        self.conv_2 = nn.Sequential(
            nn.Conv2d(C_in, C_out // 2, 1, stride=2, padding=0, bias=False)
        )
        self.bn = nn.BatchNorm2d(C_in, affine=affine)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.bn(x)
        out = torch.cat([self.conv_1(x), self.conv_2(x[:, :, 1:, 1:])], dim=1)
        out = self.activation(out)

        return out



from os import path
from setuptools import setup, find_packages
# To use consisten encodings
from codecs import open

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as readme_file:
    long_description = readme_file.read()

requirements = [
    'torch >= 1.8',
    'pyyaml',
    'easydict'
]

exec(open('bnn/version.py').read())
setup(
    name='bnn',
    version=__version__,

    description="Binarize deep convolutional neural networks using python and pytorch",
    long_description=long_description,
    long_description_content_type="text/markdown",

    # Author details
    author="Adrian Bulat",
    author_email="adrian@adrianbulat.com",
    url="https://github.com/1adrianb/binary-networks-pytorch",
    keywords=['artificial intelligence', 'convolutional neural network', 'binary networks', 'quantization', 'pytorch'],

    # Package info
    packages=find_packages(exclude=('test',)),

    install_requires=requirements,
    license='BSD',
    zip_safe=True,

    classifiers=[
        'Development Status :: 4 - Beta',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',

        # Supported python versions
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
)



from typing import Union
import torch
import torch.nn as nn

from torch.nn.common_types import _size_1_t, _size_2_t
from .. import BConfig
from .helpers import copy_paramters


class Conv1d(nn.Conv1d):
    _FLOAT_MODULE = nn.Conv1d

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: _size_1_t,
        stride: _size_1_t = 1,
        padding: Union[str, _size_1_t] = 0,
        dilation: _size_1_t = 1,
        groups: int = 1,
        bias: bool = True,
        padding_mode: str = 'zeros',
        bconfig: BConfig = None
    ) -> None:
        super(Conv2d, self).__init__(in_channels, out_channels, kernel_size,
                                     stride=stride, padding=padding, dilation=dilation,
                                     groups=groups, bias=bias, padding_mode=padding_mode)
        assert bconfig, 'bconfig is required for a binarized module'
        self.bconfig = bconfig
        self.activation_pre_process = bconfig.activation_pre_process()
        self.activation_post_process = bconfig.activation_post_process(self)
        self.weight_pre_process = bconfig.weight_pre_process()

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        input_proc = self.activation_pre_process(input)
        return self.activation_post_process(
            self._conv_forward(input_proc, self.weight_pre_process(self.weight), bias=self.bias),
            input
        )

    @classmethod
    def from_module(cls, mod: nn.Module, bconfig: BConfig = None, update: bool = False):
        assert type(mod) == cls._FLOAT_MODULE or type(mod) == cls, 'bnn.' + cls.__name__ + \
            '.from_float only works for ' + cls._FLOAT_MODULE.__name__
        if not bconfig:
            assert hasattr(mod, 'bconfig'), 'The input modele requires a predifined bconfig'
            assert mod.bconfig, 'The input modele bconfig is invalid'
            bconfig = mod.bconfig
        bnn_conv = cls(mod.in_channels, mod.out_channels, mod.kernel_size,
                       stride=mod.stride, padding=mod.padding, dilation=mod.dilation,
                       groups=mod.groups, bias=mod.bias is not None,
                       padding_mode=mod.padding_mode, bconfig=bconfig)
        bnn_conv.weight = mod.weight
        bnn_conv.bias = mod.bias

        if update:
            copy_paramters(mod, bnn_conv, bconfig)

        return bnn_conv


class Conv2d(nn.Conv2d):
    _FLOAT_MODULE = nn.Conv2d

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: _size_2_t,
        stride: _size_2_t = 1,
        padding: Union[str, _size_2_t] = 0,
        dilation: _size_2_t = 1,
        groups: int = 1,
        bias: bool = True,
        padding_mode: str = 'zeros',
        bconfig: BConfig = None
    ) -> None:
        super(Conv2d, self).__init__(in_channels, out_channels, kernel_size,
                                     stride=stride, padding=padding, dilation=dilation,
                                     groups=groups, bias=bias, padding_mode=padding_mode)
        assert bconfig, 'bconfig is required for a binarized module'
        self.bconfig = bconfig
        self.activation_pre_process = bconfig.activation_pre_process()
        self.activation_post_process = bconfig.activation_post_process(self)
        self.weight_pre_process = bconfig.weight_pre_process()

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        input_proc = self.activation_pre_process(input)
        return self.activation_post_process(
            self._conv_forward(input_proc, self.weight_pre_process(self.weight), bias=self.bias),
            input
        )

    @classmethod
    def from_module(cls, mod: nn.Module, bconfig: BConfig = None, update: bool = False):
        assert type(mod) == cls._FLOAT_MODULE or type(mod) == cls, 'bnn.' + cls.__name__ + \
            '.from_float only works for ' + cls._FLOAT_MODULE.__name__
        if not bconfig:
            assert hasattr(mod, 'bconfig'), 'The input modele requires a predifined bconfig'
            assert mod.bconfig, 'The input modele bconfig is invalid'
            bconfig = mod.bconfig
        bnn_conv = cls(mod.in_channels, mod.out_channels, mod.kernel_size,
                       stride=mod.stride, padding=mod.padding, dilation=mod.dilation,
                       groups=mod.groups, bias=mod.bias is not None,
                       padding_mode=mod.padding_mode, bconfig=bconfig)
        bnn_conv.weight = mod.weight
        bnn_conv.bias = mod.bias

        if update:
            copy_paramters(mod, bnn_conv, bconfig)

        return bnn_conv



from typing import Callable, Optional
import torch
import torch.nn as nn

from .common import conv3x3


class HBlock(nn.Module):
    def __init__(self,
                 inplanes: int,
                 planes: int,
                 stride: int = 1,
                 downsample: Optional[nn.Module] = None,
                 groups: int = 1,
                 base_width: int = 64,
                 dilation: int = 1,
                 norm_layer: Optional[Callable[..., nn.Module]] = None,
                 activation=nn.ReLU
                 ) -> None:
        super(HBlock, self).__init__()
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in HBlock")
        if stride > 1:
            raise NotImplementedError("Stride > 1 not supported in HBlock")
        self.bn1 = norm_layer(inplanes)
        self.conv1 = conv3x3(inplanes, int(planes / 2), groups=groups)
        self.bn2 = norm_layer(int(planes / 2))
        self.conv2 = conv3x3(int(planes / 2), int(planes / 4), groups=groups)
        self.bn3 = norm_layer(int(planes / 4))
        self.conv3 = conv3x3(int(planes / 4), int(planes / 4), groups=groups)

        self.act1 = activation(inpace=True) if activation == nn.ReLU else activation(num_parameters=int(planes / 2))
        self.act2 = activation(inpace=True) if activation == nn.ReLU else activation(num_parameters=int(planes / 2))
        self.act3 = activation(inpace=True) if activation == nn.ReLU else activation(num_parameters=int(planes / 4))

        self.downsample = downsample

    def forward(self, x):
        residual = x

        out1 = self.bn1(x)
        out1 = self.act1(out1)
        out1 = self.conv1(out1)

        out2 = self.bn2(out1)
        out2 = self.act2(out2)
        out2 = self.conv2(out2)

        out3 = self.bn3(out2)
        out3 = self.act3(out3)
        out3 = self.conv3(out3)

        out3 = torch.cat((out1, out2, out3), 1)

        if self.downsample is not None:
            residual = self.downsample(residual)

        out3 += residual

        return out3



import torch
import torch.nn as nn
import torch.nn.functional as F

from .. import BConfig
from .helpers import copy_paramters


class Linear(nn.Linear):
    _FLOAT_MODULE = nn.Linear

    def __init__(self, in_features: int, out_features: int, bias: bool = True,
                 bconfig: BConfig = None
                 ) -> None:
        super(Linear, self).__init__(in_features, out_features, bias)
        assert bconfig, 'bconfig is required for a binarized module'
        self.bconfig = bconfig
        self.activation_pre_process = bconfig.activation_pre_process()
        self.activation_post_process = bconfig.activation_post_process(self)
        self.weight_pre_process = bconfig.weight_pre_process()

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        input_proc = self.activation_pre_process(input)
        return self.activation_post_process(
            F.linear(input_proc, self.weight_pre_process(self.weight), self.bias),
            input
        )

    @classmethod
    def from_module(cls, mod: nn.Module, bconfig: BConfig = None, update: bool = False) -> nn.Module:
        assert type(mod) == cls._FLOAT_MODULE or type(mod) == cls, 'bnn.' + cls.__name__ + \
            '.from_float only works for ' + cls._FLOAT_MODULE.__name__
        if not bconfig:
            assert hasattr(mod, 'bconfig'), 'The input modele requires a predifined bconfig'
            assert mod.bconfig, 'The input modele bconfig is invalid'
            bconfig = mod.bconfig
        bnn_linear = cls(mod.in_features, mod.out_features, bias=mod.bias is not None,
                         bconfig=bconfig)
        bnn_linear.weight = mod.weight
        bnn_linear.bias = mod.bias

        if update:
            copy_paramters(mod, bnn_linear, bconfig)
        return bnn_linear



# model definitions based on https://github.com/pytorch/vision/blob/master/torchvision/models/resnet.py
from typing import Type, Any, Callable, Union, List, Optional
import torch
import torch.nn as nn

from .layers import BasicBlock, Bottleneck, HBlock, PreBasicBlock, PreBottleneck
from .layers import conv1x1


class DaBNNStem(nn.Module):
    def __init__(self, planes: int, norm_layer: Optional[Callable[..., nn.Module]] = None,
                 activation=nn.ReLU):
        super(DaBNNStem, self).__init__()

        self.conv1 = nn.Sequential(
            nn.Conv2d(3, planes // 2, kernel_size=3, stride=2, padding=1, bias=False),
            norm_layer(planes // 2),
            activation()
        )
        self.conv2_1 = nn.Sequential(
            nn.Conv2d(planes // 2, planes // 4, 1, 1, bias=False),
            norm_layer(planes // 4),
            activation()
        )
        self.conv2_2 = nn.Sequential(
            nn.Conv2d(planes // 4, planes // 2, kernel_size=3, stride=2, padding=1, bias=False),
            norm_layer(planes // 2),
            activation()
        )
        self.conv3 = nn.Sequential(
            nn.Conv2d(planes, planes, 1, 1, bias=False),
            norm_layer(planes),
            activation()
        )
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv1(x)
        x = torch.cat(
            [
                self.conv2_2(self.conv2_1(x)),
                self.maxpool(x)
            ],
            dim=1
        )
        x = self.conv3(x)
        return x


class ResNet(nn.Module):

    def __init__(
        self,
        block: Type[Union[BasicBlock, Bottleneck, HBlock, PreBasicBlock, PreBottleneck]],
        layers: List[int],
        num_classes: int = 1000,
        zero_init_residual: bool = False,
        groups: int = 1,
        width_per_group: int = 64,
        replace_stride_with_dilation: Optional[List[bool]] = None,
        norm_layer: Optional[Callable[..., nn.Module]] = None,
        activation: Optional[Callable[..., nn.Module]] = None,
        stem_type: str = 'basic'
    ) -> None:
        super(ResNet, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if activation is None:
            activation = nn.ReLU
        self._norm_layer = norm_layer
        self._activation = activation

        self.stem_type = stem_type
        self.inplanes = 64
        self.dilation = 1
        if replace_stride_with_dilation is None:
            # each element in the tuple indicates if we should replace
            # the 2x2 stride with a dilated convolution instead
            replace_stride_with_dilation = [False, False, False]
        if len(replace_stride_with_dilation) != 3:
            raise ValueError("replace_stride_with_dilation should be None "
                             "or a 3-element tuple, got {}".format(replace_stride_with_dilation))
        self.groups = groups
        self.base_width = width_per_group
        if stem_type == 'basic':
            self.conv1 = nn.Conv2d(3, self.inplanes, kernel_size=7, stride=2, padding=3,
                                   bias=False)
            self.bn1 = norm_layer(self.inplanes)
        elif stem_type == 'dabnn':
            self.conv1 = DaBNNStem(self.inplanes, norm_layer=norm_layer)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block, 64, layers[0])
        self.layer2 = self._make_layer(block, 128, layers[1], stride=2,
                                       dilate=replace_stride_with_dilation[0])
        self.layer3 = self._make_layer(block, 256, layers[2], stride=2,
                                       dilate=replace_stride_with_dilation[1])
        self.layer4 = self._make_layer(block, 512, layers[3], stride=2,
                                       dilate=replace_stride_with_dilation[2])
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(self.outplanes, num_classes)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

        # Zero-initialize the last BN in each residual branch,
        # so that the residual branch starts with zeros, and each residual block behaves like an identity.
        # This improves the model by 0.2~0.3% according to https://arxiv.org/abs/1706.02677
        if zero_init_residual:
            for m in self.modules():
                if isinstance(m, Bottleneck):
                    nn.init.constant_(m.bn3.weight, 0)  # type: ignore[arg-type]
                elif isinstance(m, BasicBlock):
                    nn.init.constant_(m.bn2.weight, 0)  # type: ignore[arg-type]

    def _make_layer(self, block: Type[Union[BasicBlock, Bottleneck]], planes: int, blocks: int,
                    stride: int = 1, dilate: bool = False) -> nn.Sequential:
        norm_layer = self._norm_layer
        downsample = None
        previous_dilation = self.dilation
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.AvgPool2d(kernel_size=stride, stride=stride, ceil_mode=True, count_include_pad=False),
                conv1x1(self.inplanes, planes * block.expansion, stride=1),
                norm_layer(planes * block.expansion),
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample, self.groups,
                            self.base_width, previous_dilation, norm_layer, activation=self._activation))
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(block(self.inplanes, planes, groups=self.groups,
                                base_width=self.base_width, dilation=self.dilation,
                                norm_layer=norm_layer, activation=self._activation))
        self.outplanes = planes

        return nn.Sequential(*layers)

    def _forward_impl(self, x: torch.Tensor) -> torch.Tensor:
        # See note [TorchScript super()]
        x = self.conv1(x)
        if self.stem_type == 'basic':
            x = self.bn1(x)
            x = self.relu(x)
            x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        x = self.fc(x)

        return x

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self._forward_impl(x)


def _resnet(
    arch: str,
    block: Type[Union[BasicBlock, Bottleneck, HBlock, PreBottleneck, PreBasicBlock]],
    layers: List[int],
    **kwargs: Any
) -> ResNet:
    model = ResNet(block, layers, **kwargs)
    return model


def resnet18(block_type: Optional[Type[Union[BasicBlock, Bottleneck, HBlock,
                                             PreBottleneck, PreBasicBlock]]] = None, **kwargs: Any) -> ResNet:
    r"""ResNet-18 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _resnet('resnet18', BasicBlock if block_type is None else block_type, [2, 2, 2, 2],
                   **kwargs)


def resnet34(block_type: Optional[Type[Union[BasicBlock, Bottleneck, HBlock,
                                             PreBottleneck, PreBasicBlock]]] = None, **kwargs: Any) -> ResNet:
    r"""ResNet-34 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _resnet('resnet34', BasicBlock if block_type is None else block_type, [3, 4, 6, 3],
                   **kwargs)


def resnet50(block_type: Optional[Type[Union[BasicBlock, Bottleneck, HBlock,
                                             PreBottleneck, PreBasicBlock]]] = None, **kwargs: Any) -> ResNet:
    r"""ResNet-50 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _resnet('resnet50', Bottleneck if block_type is None else block_type, [3, 4, 6, 3],
                   **kwargs)
